# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Enrico Faulhaber <enrico.faulhaber@frm2.tum.de>
#
# *****************************************************************************

"""Basic test suite for the devices in the testplc."""

import time

import pytest

from zapf import ApiError
from zapf.device import typecode_description
from zapf.spec import ParamCMDs, PLCStatus


def test_readable_device(plc_readable_device):
    """test value/status"""
    # zapf-device, TYPCODE_MAP entry
    dev, _tcme = plc_readable_device

    status = dev.read_status()
    state = status.STATE

    assert state in [PLCStatus.IDLE, PLCStatus.DISABLED,
                     PLCStatus.WARN, PLCStatus.START,
                     PLCStatus.BUSY, PLCStatus.STOP, PLCStatus.ERROR]
    assert dev.name == typecode_description(dev.typecode)

    value = dev.read_value()
    limits = dev.get_limits()
    assert limits[0] <= value <= limits[1]

    if not dev.target_addr:
        with pytest.raises(ApiError):
            dev.read_target()
        with pytest.raises(ApiError):
            dev.change_target(0)


def test_target_device(plc_target_device):
    """test target, moving around"""
    # zapf-device, TYPCODE_MAP entry
    dev, _tcme = plc_target_device

    status = dev.read_status()

    target = dev.read_target()
    limits = dev.get_limits()
    assert limits[0] <= target <= limits[1]

    # check precondition before starting a movement
    assert status.STATE not in [PLCStatus.START, PLCStatus.BUSY,
                                PLCStatus.STOP]
    assert status.STATE in [PLCStatus.IDLE, PLCStatus.WARN]

    value = dev.read_value()
    target = value + 1
    dev.change_target(target)
    assert dev.read_target() == target

    timesout = time.time() + 1
    # now check for end of movement
    while True:
        status = dev.read_status()
        value = dev.read_value()
        target = dev.read_target()
        assert status.STATE in [PLCStatus.START, PLCStatus.BUSY,
                                PLCStatus.IDLE, PLCStatus.WARN]
        if value != target:
            assert status.STATE == PLCStatus.BUSY
        if status.STATE != PLCStatus.BUSY:
            assert value == target
            return
        assert time.time() < timesout


def test_vector_devices(plc_vector_device):
    """test vector value/status/target/moving around"""
    # zapf-device, TYPCODE_MAP entry
    dev, tcme = plc_vector_device

    has_target, num_values = tcme.has_target, tcme.num_values

    status = dev.read_status()
    state = status.STATE

    value = dev.read_value()
    assert len(value) == num_values

    limits = dev.get_limits()
    for v in value:
        assert limits[0] <= v <= limits[1]

    if not has_target:
        with pytest.raises(ApiError):
            dev.read_target()
        with pytest.raises(ApiError):
            dev.change_target(value)
        return

    target = dev.read_target()
    for t in target:
        assert limits[0] <= t <= limits[1]
    if target == value:
        assert state in [PLCStatus.IDLE, PLCStatus.DISABLED,
                         PLCStatus.WARN]
    else:
        assert state in [PLCStatus.DISABLED, PLCStatus.START,
                         PLCStatus.BUSY, PLCStatus.STOP, PLCStatus.ERROR]


def test_flat_device(plc_flat_device):
    """test flat parameter access"""
    # zapf-device, TYPCODE_MAP entry
    dev, tcme = plc_flat_device

    # check param access to UserMin
    v = dev.get_param('UserMin')[1]
    dev.set_param('UserMin', v+1)
    assert v+1 == dev.get_param('UserMin')[1]
    dev.set_param('UserMin', v)

    assert tcme.num_params == len(dev.list_params())

    # check param access of all other params
    for p in dev.list_params():
        print(p)
        if p == 'UserMin':
            continue
        v = dev.get_param(p)[1]
        dev.set_param(p, v+1)
        assert v+1 == dev.get_param(p)[1]
        dev.set_param(p, v)


def test_paramif(plc_param_device):
    """test paramif access + funcs"""
    # zapf-device, TYPCODE_MAP entry
    dev, _tcme = plc_param_device

    # testplc implements a few parameters:
    # readable: 1-3, 32-38, 40, 43, 44, 51-53, 55-65, 68-70
    # writeable: 32-37, 40, 51-53, 56, 57, 60, 61, 69, 70
    # executable: 128, 133, 137, 142
    # 128 and 133 almost instantly go from BUSY to DONE, rest is ERR_RETRY

    # check state machine
    dev.update_param_sm()
    assert dev.param_sm.CMD != 0

    # basic lowlevel tests

    # set a known paramvalue != 0 first
    dev.set_param_value(1, 1)
    # initiate reading parameter 1 (should result in value 0)
    dev.set_param_cmd(ParamCMDs.DO_READ, 0, 1)
    assert dev.wait_sm_available()
    assert dev.param_sm.CMD == ParamCMDs.DONE
    assert dev.get_param_value(1) == 0

    # try to read param 127 (should fail with NO_IDX)
    assert dev.wait_sm_available()
    dev.set_param_cmd(ParamCMDs.DO_READ, 0, 127)
    assert dev.wait_sm_available()
    assert dev.param_sm.CMD == ParamCMDs.ERR_NO_IDX

    # try to write param 1 (should fail with RO)
    assert dev.wait_sm_available()
    dev.set_param_value_and_cmd(ParamCMDs.DO_WRITE, 0, 1, 42)
    assert dev.wait_sm_available()
    assert dev.param_sm.CMD == ParamCMDs.ERR_RO

    # try to exec func 142 (should fail with RETRY)
    assert dev.wait_sm_available()
    dev.set_param_cmd(ParamCMDs.DO_WRITE, 0, 142)
    assert dev.wait_sm_available()
    assert dev.param_sm.CMD == ParamCMDs.ERR_RETRY

    # try to 'BUSY' param 1 (should succeed with DONE)
    assert dev.wait_sm_available()
    dev.set_param_value_and_cmd(ParamCMDs.BUSY, 0, 1, 42)
    assert dev.wait_sm_available()
    assert dev.param_sm.CMD == ParamCMDs.DONE
    assert dev.get_param_value(1) != 42

    # higher level tests
    if 'UserMax' in dev.list_params():
        v = dev.get_param('UserMax')[1]
        dev.set_param('UserMax', v + 1)
        assert dev.get_param('UserMax')[1] == v + 1

    if dev.list_funcs():
        dev.exec_func(dev.list_funcs()[0])



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Interaction with the indexer."""

import time

from zapf import SpecError, spec


class Indexer:
    def __init__(self, io, log):
        self.io = io
        self.log = log
        self.magicstr = None
        self.indexer_addr = None
        self.indexer_size = None
        self.num_devices = 0
        self.plc_name = None
        self.plc_version = None
        self.plc_author = None

    def detect_plc(self, extended=True):
        """Checks if the HW follows the spec and which MAGIC number it has.

        Afterwards, checks the indexer and reads extended meta information
        about the PLC if wanted.
        """
        if self.indexer_addr and self.indexer_size and self.plc_author:
            return

        # check MAGIC
        magic = self.io.detect_magic()
        magicstr = ('%.2f' % round(magic, 4)).replace('.', '_')
        if magicstr not in spec.SUPPORTED_MAGICS:
            raise SpecError(f'magic {magicstr} is not supported by this client')
        self.magicstr = magicstr

        # read indexer offset from memory address 4
        addr = self.io.read_u16(spec.OFFSET_ADDR)
        if addr < 6 or addr & 1:
            raise SpecError('indexer offset %s is invalid' % addr)
        self.indexer_addr = addr

        # query indexer size
        size = self.query_word(spec.INDEXER_DEV, spec.INFO_SIZE)
        if size < 22 or size > 66 or size & 1:
            raise SpecError('indexer size %s is invalid' % size)
        self.indexer_size = size

        # query indexer size and offset again, through the info struct,
        # and ensure consistency
        info = self.query_infostruct(spec.INDEXER_DEV)
        if info[0] + info[1] + info[2] != 0:
            if info[0] != 0 or info[1] != size or info[2] != addr:
                raise SpecError('indexer information from infostruct does not '
                                'match with OFFSET or size')

            # total number of devices can be given in flags
            self.num_devices = info[4] & 0xFF

        # query firmware information
        if extended:
            self.plc_name = self.query_string(spec.INDEXER_DEV, spec.INFO_NAME)
            self.plc_version = self.query_string(spec.INDEXER_DEV,
                                                 spec.INFO_VERSION)
            author1 = self.query_string(spec.INDEXER_DEV, spec.INFO_AUTHOR1)
            author2 = self.query_string(spec.INDEXER_DEV, spec.INFO_AUTHOR2)
            self.plc_author = (author1 or 'Anonymous') + '\n' + author2

    # lowlevel methods to query the indexer

    def query_infostruct(self, devnum):
        # fields in the infostruct: typecode, size, addr, unitcode, unitexp,
        # flags, absmin, absmax, the rest is the name
        result = self.query_data(devnum, spec.INFO_STRUCT,
                                 'HHHBbIII%ds' % (self.indexer_size - 22))
        # convert min/max; these are floats, but we need to potentially word-
        # swap them
        absmin = self.io.float_from_dword(result[6])
        absmax = self.io.float_from_dword(result[7])
        # convert unit to a string
        unit = self.convert_unit(*result[3:5])
        # only use the name if it has a trailing null byte, so we can be
        # sure it was fully transferred in the reduced space at the end
        name_parts = result[-1].partition(b'\0')
        name = name_parts[0].decode('latin1') if name_parts[1] else ''
        return result[:3] + (unit, result[5], absmin, absmax, name)

    def query_word(self, devnum, infotype):
        return self.query_data(devnum, infotype, 'H')[0]

    def query_unit(self, devnum, infotype):
        return self.convert_unit(*self.query_data(devnum, infotype, 'Bb'))

    def query_string(self, devnum, infotype):
        result = self.query_bytes(devnum, infotype)
        return result.partition(b'\0')[0].decode('latin1')

    def query_bitmap(self, devnum, infotype):
        return [grp*8 + bit
                for grp, byte in enumerate(self.query_bytes(devnum, infotype))
                for bit in range(8)
                if byte & (1 << bit)]

    def query_bytes(self, devnum, infotype):
        return self.query_data(devnum, infotype,
                               '%ds' % (self.indexer_size - 2))[0]

    # even lower level methods to do an indexer transaction

    def query_data(self, devnum, infotype, fmt):
        request = infotype << 8 | devnum
        self.io.write_u16s(self.indexer_addr, [request])
        for i in range(32):
            reply = self.io.read_fmt(self.indexer_addr, 'H' + fmt)
            if reply[0] == request | 0x8000:
                return reply[1:]
            # refresh the query if it got overwritten by a different host
            if reply[0] & 0x7fff != request:
                self.io.write_u16s(self.indexer_addr, [request])
            time.sleep(0.001*i*i)
        raise SpecError('indexer not responding in time!')

    def convert_unit(self, code, exponent):
        if (code, exponent) in spec.UNIT_SPECIAL:
            return spec.UNIT_SPECIAL[code, exponent]
        try:
            unit = spec.UNIT_CODES[code]
        except IndexError:
            unit = 'unit'
        if '^' in unit:
            return '10^%d %s' % (exponent, unit)
        return spec.UNIT_EXPONENT.get(exponent, '10^%d ' % exponent) + unit



#  -*- coding: utf-8 -*-
# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************


class NumProxy:

    def __add__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value + other

    def __sub__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value - other

    def __mul__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value * other

    def __floordiv__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value // other

    def __truediv__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value / other

    def __mod__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value % other

    def __and__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value & other

    def __or__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value | other

    def __xor__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value ^ other

    def __rshift__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value >> other

    def __lshift__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value << other

    def __lt__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value < other

    def __le__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value <= other

    def __ge__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value >= other

    def __gt__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value > other

    def __eq__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value == other

    def __ne__(self, other):
        if isinstance(other, NumProxy):
            other = other.value
        return self.value != other

    def __bool__(self):
        return self.value.__bool__()

    def __neg__(self):
        return self.value.__neg__()

    def __pos__(self):
        return self.value.__pos__()

    def __abs__(self):
        return self.value.__abs__()

    def __invert__(self):
        return self.value.__invert__()

    def __index__(self):
        return self.value.__index__()

    def __int__(self):
        return self.value.__int__()

    def __float__(self):
        return self.value.__float__()

    def __hash__(self):
        return self.value.__hash__()



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Douglas Creager <dcreager@dcreager.net>
#   This file is placed into the public domain.
#
# *****************************************************************************

import os.path
from subprocess import PIPE, Popen

__all__ = ['get_version']

RELEASE_VERSION_FILE = os.path.join(os.path.dirname(__file__),
                                    'RELEASE-VERSION')
GIT_REPO = os.path.join(os.path.dirname(__file__), '..', '.git')


def translate_version(ver):
    ver = ver.lstrip('v').rsplit('-', 2)
    return '%s.post%s+%s' % tuple(ver) if len(ver) == 3 else ver[0]


def get_git_version(abbrev=4):
    try:
        with Popen(['git', '--git-dir=%s' % GIT_REPO,
                    'describe', '--abbrev=%d' % abbrev],
                   stdout=PIPE, stderr=PIPE) as p:
            stdout, _stderr = p.communicate()
        return translate_version(stdout.strip().decode('utf-8', 'ignore'))
    except Exception:
        return None


def read_release_version():
    try:
        with open(RELEASE_VERSION_FILE) as f:
            return f.readline().strip()
    except Exception:
        return None


def write_release_version(version):
    with open(RELEASE_VERSION_FILE, 'w') as f:
        f.write("%s\n" % version)


def get_version(abbrev=4):
    # determine the version from git and from RELEASE-VERSION
    git_version = get_git_version(abbrev)
    release_version = read_release_version()

    # if we have a git version, it is authoritative
    if git_version:
        if git_version != release_version:
            write_release_version(git_version)
        return git_version
    elif release_version:
        return release_version
    else:
        raise ValueError('Cannot find a version number - make sure that '
                         'git is installed or a RELEASE-VERSION file is '
                         'present!')


if __name__ == "__main__":
    print(get_version())



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Implements communication to a simulated PLC running in-process."""

import threading
import time

from zapf import ApiError, CommError
from zapf.proto import Protocol
from zapf.simulator import runtime


class SimProtocol(Protocol):
    OFFSETS = [0x10000]

    def __init__(self, url, log):
        if callable(url):
            self._plcfunc = url
        else:
            try:
                if not url.startswith('sim://'):
                    raise Exception('URL should start with sim://')
                with open(url[6:]) as fp:
                    code = fp.read()
                globs = {}
                # TODO: execute with the correct filename to get
                # filename  and code in tracebacks
                exec(code, globs)  # pylint: disable=exec-used
                self._plcfunc = globs['Main']
            except Exception as err:
                raise ApiError('invalid sim address, must be '
                               'sim:///path/to/file.py '
                               'with a Main() function') from err

        self._cond = threading.Condition()
        self._thread = None
        self._stopflag = False
        self._exception = None

        Protocol.__init__(self, url, log)

    def connect(self):
        if self._thread and self._thread.is_alive():
            return
        self._stopflag = False
        self._thread = threading.Thread(target=self._plc_thread, daemon=True,
                                        name=f'SimPLC {self.url}')
        self._thread.start()
        self._connect_callback(True)
        self.connected = True

    def disconnect(self):
        if self._thread:
            self._stopflag = True
            self._thread.join()
        self._connect_callback(False)
        self.connected = False

    def read(self, addr, length):
        with self._cond:
            self._cond.wait()
            try:
                return runtime.mem.read(self.offset + addr, length)
            except RuntimeError as err:
                raise CommError(f'reading {addr}({length}): {err}') from err

    def write(self, addr, data):
        with self._cond:
            self._cond.wait()
            try:
                runtime.mem.write(self.offset + addr, data)
            except RuntimeError as err:
                raise CommError(f'writing {addr}({len(data)}): {err}') from err

    def _plc_thread(self):
        while not self._stopflag:
            with self._cond:
                try:
                    if not self._exception:
                        self._plcfunc()
                except Exception as err:
                    self.log.exception('error in PLC simulation thread, '
                                       'aborting simulation')
                    self._exception = err
                self._cond.notify()
            time.sleep(.0005)



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Basic device abstraction for PILS devices."""

import time
from collections import namedtuple
from struct import calcsize

from zapf import ApiError, SpecError
from zapf.spec import FIRST_FLOAT_PARAM, FLOAT32_MAX, FLOAT64_MAX, ParamCMDs, \
    ParamControl, Parameters, PLCStatus, StatusStruct
from zapf.util import UncasedMap

START16 = PLCStatus.START << 12
START32 = PLCStatus.START << 28


class Device:
    @classmethod
    def class_for(cls, typecode):
        return TYPECODE_MAP.get(typecode, (None,))[0]

    valuetype = None

    def __init__(self, number, name, addr, typecode, info, io, log):
        self.number = number
        self.name = name
        self.typecode = typecode
        self.device_kind = typecode >> 8
        self.total_size = 2 * (typecode & 0xff)
        self.info = info
        self.params = info['params']
        self.funcs = info['funcs']

        typeinfo = TYPECODE_MAP[typecode]
        self.value_fmt = typeinfo.value_fmt
        self.value_size = calcsize(self.value_fmt)
        self.status_size = typeinfo.status_size
        self.num_values = typeinfo.num_values
        self.num_params = typeinfo.num_params

        self.addr = addr
        next_addr = addr + self.num_values*self.value_size
        if typeinfo.has_target:
            self.target_addr = next_addr
            next_addr += self.num_values*self.value_size
        else:
            self.target_addr = None
        if self.status_size:
            self.status_addr = next_addr
        else:
            self.status_addr = None
        if typeinfo.has_pctrl:
            self.pctrl_addr = self.status_addr + self.status_size
        else:
            self.pctrl_addr = None
        if typeinfo.num_params:
            self.param_addr = self.status_addr + self.status_size + 2
        else:
            self.param_addr = None

        self.log = log
        self.io = io
        self.init()
        self.io.register_cache_range(self.addr, self.total_size)

    def read_status(self):
        if self.status_addr is None:
            if self.target_addr is None:
                return StatusStruct(PLCStatus.IDLE << 12)
            if self.read_value() == self.read_target():
                return StatusStruct(PLCStatus.IDLE << 12)
            return StatusStruct(PLCStatus.BUSY << 12)
        if self.status_size == 2:
            return StatusStruct(self.io.read_u16(self.status_addr))
        return StatusStruct(self.io.read_u32(self.status_addr) >> 16)

    def read_extended_status(self):
        if self.status_addr is None:
            return (StatusStruct(PLCStatus.IDLE << 12), 0, 0)
        if self.status_size == 2:
            return (StatusStruct(self.io.read_u16(self.status_addr)), 0, 0)
        elif self.status_size == 4:
            value = self.io.read_u32(self.status_addr)
            return (StatusStruct(value >> 16), value & 0xFFFF, 0)
        elif self.status_size == 6:
            value, err_id = self.io.read_fmt(self.status_addr, 'IH')
            return (StatusStruct(value >> 16), value & 0xFFFF, err_id)
        raise SpecError('invalid status_size')

    def change_status(self, initial_states=(), final_state=0):
        if self.status_addr is None:
            return False
        if initial_states:
            status = self.read_status()
            if status.STATE not in initial_states:
                return False
        else:
            status = StatusStruct()
        status.STATE = final_state
        if self.status_size == 2:
            self.io.write_u16(self.status_addr, int(status))
        else:
            self.io.write_u32(self.status_addr, int(status) << 16)
        return True

    # to implement:

    def init(self):
        raise NotImplementedError

    def get_limits(self):
        raise NotImplementedError

    def read_value(self):
        raise NotImplementedError

    def read_target(self):
        raise ApiError('reading target of a read-only device')

    def change_target(self, value):
        raise ApiError('writing target of a read-only device')

    def list_params(self):
        return []

    def get_param(self, name):
        raise ApiError('reading parameter of a device without params')

    def set_param(self, name, value):
        raise ApiError('writing parameter of a device without params')

    def list_funcs(self):
        return []

    def exec_func(self, name, value=None):
        raise ApiError('executing function of a device without functions')


class DiscreteDevice(Device):
    def init(self):
        pass

    def get_limits(self):
        if self.value_size == 2:
            return (-2**15, 2**15 - 1)
        elif self.value_size == 4:
            return (-2**31, 2**31 - 1)
        elif self.value_size == 8:
            return (-2**63, 2**63 - 1)
        raise SpecError('invalid value_size')

    def read_value(self):
        return self.io.read_fmt(self.addr, self.value_fmt)[0]

    def read_target(self):
        if self.target_addr is None:
            raise ApiError('reading target of a read-only device')
        return self.io.read_fmt(self.target_addr, self.value_fmt)[0]

    def change_target(self, value):
        if self.target_addr is None:
            raise ApiError('writing target of a read-only device')
        if self.status_addr is None:
            self.io.write_fmt(self.target_addr, self.value_fmt, value)
        elif self.value_size == 2:
            self.io.write_fmt(self.target_addr, self.value_fmt + 'H',
                              value, START16)
        else:
            self.io.write_fmt(self.target_addr, self.value_fmt + 'I',
                              value, START32)


class SimpleDiscreteIn(DiscreteDevice):
    pass


class SimpleDiscreteOut(DiscreteDevice):
    pass


class DiscreteIn(DiscreteDevice):
    pass


class DiscreteOut(DiscreteDevice):
    pass


class Keyword(DiscreteDevice):
    def init(self):
        self.target_addr = self.addr

    def get_limits(self):
        if self.value_size == 2:
            return (0, 2**16 - 1)
        elif self.value_size == 4:
            return (0, 2**32 - 1)
        elif self.value_size == 8:
            return (0, 2**64 - 1)
        raise SpecError('invalid value_size')


class StatusWord(Keyword):
    def init(self):
        Keyword.init(self)
        self.status_addr = self.addr


class AnalogDevice(Device):
    def init(self):
        pass

    def get_limits(self):
        if self.value_size == 4:
            return (-FLOAT32_MAX, FLOAT32_MAX)
        elif self.value_size == 8:
            return (-FLOAT64_MAX, FLOAT64_MAX)
        raise SpecError('invalid value_size')

    def read_value(self):
        if self.value_size == 4:
            return self.io.read_f32(self.addr)
        elif self.value_size == 8:
            return self.io.read_f64(self.addr)
        raise SpecError('invalid value_size')

    def read_target(self):
        if self.target_addr is None:
            raise ApiError('reading target of a read-only device')
        if self.value_size == 4:
            return self.io.read_f32(self.target_addr)
        elif self.value_size == 8:
            return self.io.read_f64(self.target_addr)
        raise SpecError('invalid value_size')

    def change_target(self, value):
        if self.target_addr is None:
            raise ApiError('writing target of a read-only device')
        if self.status_addr is None:
            if self.value_size == 4:
                self.io.write_f32(self.target_addr, value)
            elif self.value_size == 8:
                self.io.write_f64(self.target_addr, value)
        elif self.value_size == 4:
            if self.status_size == 2:
                self.io.write_f32_u16(self.target_addr, value, START16)
            else:
                self.io.write_f32_u32(self.target_addr, value, START32)
        elif self.value_size == 8:
            self.io.write_fmt(self.target_addr, 'dI', value, START32)


class SimpleAnalogIn(AnalogDevice):
    pass


class SimpleAnalogOut(AnalogDevice):
    pass


class AnalogIn(AnalogDevice):
    pass


class AnalogOut(AnalogDevice):
    pass


class RealValue(AnalogDevice):
    def init(self):
        self.target_addr = self.addr


class FlatParams:
    def init(self):
        if self.num_params != len(self.params):
            raise SpecError('mismatch between parameter count between '
                            'typecode and parameter indices from indexer '
                            f'({self.num_params}/{len(self.params)})')
        # sort params from indexer by index
        pars_indices = sorted((Parameters[par], par) for par in self.params)
        self.param_map = UncasedMap(*(
            (par[1], (self.param_addr + i * self.value_size, par[0]))
            for (i, par) in enumerate(pars_indices)
        ))

    def list_params(self):
        return self.param_map.init_keys()

    def get_param(self, name):
        (addr, idx) = self.param_map.get(name, (None, None))
        if addr:
            if idx < FIRST_FLOAT_PARAM:
                if self.value_size == 4:
                    return ParamCMDs.DONE, self.io.read_u32(addr)
                else:
                    return ParamCMDs.DONE, self.io.read_u64(addr)
            else:
                if self.value_size == 4:
                    return ParamCMDs.DONE, self.io.read_f32(addr)
                else:
                    return ParamCMDs.DONE, self.io.read_f64(addr)
        return ParamCMDs.ERR_NO_IDX, None

    def set_param(self, name, value):
        (addr, idx) = self.param_map.get(name, (None, None))
        if addr:
            if idx < FIRST_FLOAT_PARAM:
                if self.value_size == 4:
                    self.io.write_u32(addr, int(value))
                    return ParamCMDs.DONE, self.io.read_u32(addr)
                else:
                    self.io.write_u64(addr, int(value))
                    return ParamCMDs.DONE, self.io.read_u64(addr)
            else:
                if self.value_size == 4:
                    self.io.write_f32(addr, value)
                    return ParamCMDs.DONE, self.io.read_f32(addr)
                else:
                    self.io.write_f64(addr, value)
                    return ParamCMDs.DONE, self.io.read_f64(addr)
        return ParamCMDs.ERR_NO_IDX, None


class FlatIn(FlatParams, AnalogDevice):
    pass


class FlatOut(FlatParams, AnalogDevice):
    pass


class ParamInterface:
    param_timeout = 1

    def init(self):
        self.param_map = UncasedMap(*((p, Parameters[p]) for p in self.params))
        self.func_map = UncasedMap(*((p, Parameters[p]) for p in self.funcs))
        self.param_sm = ParamControl()

    def update_param_sm(self):
        self.param_sm(self.io.read_u16(self.pctrl_addr))

    def wait_sm_available(self):
        self.update_param_sm()
        timesout = time.time() + self.param_timeout
        while not self.param_sm.available:
            self.update_param_sm()
            if time.time() > timesout:
                return False
        return True

    def set_param_value(self, idx, value):
        if idx < FIRST_FLOAT_PARAM:
            if self.value_size == 4:
                self.io.write_u32(self.param_addr, int(value))
            else:
                self.io.write_u64(self.param_addr, int(value))
        else:
            if self.value_size == 4:
                self.io.write_f32(self.param_addr, value)
            else:
                self.io.write_f64(self.param_addr, value)

    def set_param_value_and_cmd(self, cmd, sub, idx, value):
        self.param_sm.CMD = cmd
        self.param_sm.SUB = sub
        self.param_sm.IDX = idx
        if idx < FIRST_FLOAT_PARAM:
            if self.value_size == 4:
                self.io.write_fmt(self.pctrl_addr, 'HI',
                                  int(self.param_sm), int(value))
            else:
                self.io.write_fmt(self.pctrl_addr, 'HQ',
                                  int(self.param_sm), int(value))
        else:
            if self.value_size == 4:
                self.io.write_u16_f32(self.pctrl_addr,
                                      int(self.param_sm), value)
            else:
                self.io.write_fmt(self.pctrl_addr, 'Hd',
                                  int(self.param_sm), value)

    def set_param_cmd(self, cmd, sub, idx):
        self.param_sm.CMD = cmd
        self.param_sm.SUB = sub
        self.param_sm.IDX = idx
        self.io.write_u16(self.pctrl_addr, int(self.param_sm))

    def get_param_value(self, idx):
        if idx < FIRST_FLOAT_PARAM:
            if self.value_size == 4:
                return self.io.read_u32(self.param_addr)
            else:
                return self.io.read_u64(self.param_addr)
        else:
            if self.value_size == 4:
                return self.io.read_f32(self.param_addr)
            else:
                return self.io.read_f64(self.param_addr)

    def list_params(self):
        return self.param_map.init_keys()

    def get_param(self, name, subdev=0):
        idx = self.param_map.get(name)
        if idx:
            if self.wait_sm_available():
                self.set_param_cmd(ParamCMDs.DO_READ, subdev, idx)
                # now wait until setting parameter is finished and
                # return read-back-value
                if self.wait_sm_available():
                    return self.param_sm.CMD, self.get_param_value(idx)
            return ParamCMDs.ERR_RETRY, None
        return ParamCMDs.ERR_NO_IDX, None

    def set_param(self, name, value, subdev=0):
        idx = self.param_map.get(name)
        if idx:
            if self.wait_sm_available():
                self.set_param_value_and_cmd(ParamCMDs.DO_WRITE,
                                             subdev, idx, value)
                # now wait until setting parameter is finished and
                # return read-back-value
                if self.wait_sm_available():
                    return self.param_sm.CMD, self.get_param_value(idx)
            return ParamCMDs.ERR_RETRY, None
        return ParamCMDs.ERR_NO_IDX, None

    def list_funcs(self):
        return self.func_map.init_keys()

    def exec_func(self, name, value=None, subdev=0):
        idx = self.func_map.get(name)
        if idx:
            self.update_param_sm()
            if self.param_sm.CMD == ParamCMDs.BUSY and \
               self.param_sm.SUB == subdev and \
               self.param_sm.IDX == idx:
                self.set_param_value(idx, value)
                if self.wait_sm_available():
                    return self.param_sm.CMD, self.get_param_value(idx)
                return self.param_sm.CMD, self.get_param_value(idx)  # still BUSY
            if self.wait_sm_available():
                self.set_param_value_and_cmd(ParamCMDs.DO_WRITE,
                                             subdev, idx, value or 0)
                # now wait until setting parameter is finished and return
                # read-back-value
                if self.wait_sm_available():
                    return self.param_sm.CMD, self.get_param_value(idx)
                return self.param_sm.CMD, self.get_param_value(idx)  # still BUSY
            return ParamCMDs.ERR_RETRY, None
        return ParamCMDs.ERR_NO_IDX, None


class ParamIn(ParamInterface, AnalogDevice):
    pass


class ParamOut(ParamInterface, AnalogDevice):
    pass


class VectorDevice(Device):
    def init(self):
        pass

    def get_limits(self):
        if self.value_size == 4:
            return (-FLOAT32_MAX, FLOAT32_MAX)
        elif self.value_size == 8:
            return (-FLOAT64_MAX, FLOAT64_MAX)
        raise SpecError('invalid value_size')

    def read_value(self):
        if self.value_size == 4:
            return self.io.read_f32s(self.addr, self.num_values)
        elif self.value_size == 8:
            return self.io.read_f64s(self.addr, self.num_values)
        raise SpecError('invalid value_size')


class VectorIn(ParamInterface, VectorDevice):
    pass


class VectorOut(ParamInterface, VectorDevice):
    def read_target(self):
        if self.value_size == 4:
            return self.io.read_f32s(self.target_addr, self.num_values)
        elif self.value_size == 8:
            return self.io.read_f64s(self.target_addr, self.num_values)
        raise SpecError('invalid value_size')

    def change_target(self, value):
        if self.value_size == 4:
            self.io.write_f32s_u16(self.target_addr, value, START16)
        elif self.value_size == 8:
            self.io.write_f64s_u32(self.target_addr, value, START32)


# Note: has_target means that a target field is there, not that the device
# is considered read-only.
Type = namedtuple(
    'Type',
    'devcls value_fmt num_values has_target status_size num_params has_pctrl')


TYPECODE_MAP = {
    0x1201: Type(SimpleDiscreteIn,  'h', 1, False, 0, 0, False),
    0x1202: Type(SimpleDiscreteIn,  'i', 1, False, 0, 0, False),
    0x1204: Type(SimpleDiscreteIn,  'q', 1, False, 0, 0, False),
    0x1302: Type(SimpleAnalogIn,    'f', 1, False, 0, 0, False),
    0x1304: Type(SimpleAnalogIn,    'd', 1, False, 0, 0, False),
    0x1401: Type(Keyword,           'H', 1, False, 0, 0, False),
    0x1402: Type(Keyword,           'I', 1, False, 0, 0, False),
    0x1404: Type(Keyword,           'Q', 1, False, 0, 0, False),
    0x1502: Type(RealValue,         'f', 1, False, 0, 0, False),
    0x1504: Type(RealValue,         'd', 1, False, 0, 0, False),
    0x1602: Type(SimpleDiscreteOut, 'h', 1, True,  0, 0, False),
    0x1604: Type(SimpleDiscreteOut, 'i', 1, True,  0, 0, False),
    0x1608: Type(SimpleDiscreteOut, 'q', 1, True,  0, 0, False),
    0x1704: Type(SimpleAnalogOut,   'f', 1, True,  0, 0, False),
    0x1708: Type(SimpleAnalogOut,   'd', 1, True,  0, 0, False),
    0x1801: Type(StatusWord,        'H', 1, False, 2, 0, False),
    0x1802: Type(StatusWord,        'I', 1, False, 4, 0, False),
    0x1a02: Type(DiscreteIn,        'h', 1, False, 2, 0, False),
    0x1a04: Type(DiscreteIn,        'i', 1, False, 4, 0, False),
    0x1a08: Type(DiscreteIn,        'q', 1, False, 6, 0, False),
    0x1b03: Type(AnalogIn,          'f', 1, False, 2, 0, False),
    0x1b04: Type(AnalogIn,          'f', 1, False, 4, 0, False),
    0x1b08: Type(AnalogIn,          'd', 1, False, 6, 0, False),
    0x1e03: Type(DiscreteOut,       'h', 1, True,  2, 0, False),
    0x1e06: Type(DiscreteOut,       'i', 1, True,  4, 0, False),
    0x1e0c: Type(DiscreteOut,       'q', 1, True,  6, 0, False),
    0x1f05: Type(AnalogOut,         'f', 1, True,  2, 0, False),
    0x1f06: Type(AnalogOut,         'f', 1, True,  4, 0, False),
    0x1f0c: Type(AnalogOut,         'd', 1, True,  6, 0, False),

    0x4006: Type(ParamIn,           'f', 1, False, 2, 1, True),
    0x400c: Type(ParamIn,           'd', 1, False, 6, 1, True),
    0x5008: Type(ParamOut,          'f', 1, True,  2, 1, True),
    0x5010: Type(ParamOut,          'd', 1, True,  6, 1, True),
}

for n in range(16):
    fin32  = 0x2000 | (n << 8) | (6 + 2*n)
    fin64  = 0x2000 | (n << 8) | (12 + 4*n)
    fout32 = 0x3000 | (n << 8) | (8 + 2*n)
    fout64 = 0x3000 | (n << 8) | (16 + 4*n)
    TYPECODE_MAP[fin32]  = Type(FlatIn,  'f', 1, False, 2, n+1, False)
    TYPECODE_MAP[fin64]  = Type(FlatIn,  'd', 1, False, 6, n+1, False)
    TYPECODE_MAP[fout32] = Type(FlatOut, 'f', 1, True,  2, n+1, False)
    TYPECODE_MAP[fout64] = Type(FlatOut, 'd', 1, True,  6, n+1, False)

    if n == 0:
        continue

    vin32  = 0x4000 | (n << 8) | (6 + 2*n)
    vin64  = 0x4000 | (n << 8) | (12 + 4*n)
    vout32 = 0x5000 | (n << 8) | (8 + 4*n)
    vout64 = 0x5000 | (n << 8) | (16 + 8*n)
    TYPECODE_MAP[vin32]  = Type(VectorIn,  'f', n+1, False, 2, 1, True)
    TYPECODE_MAP[vin64]  = Type(VectorIn,  'd', n+1, False, 6, 1, True)
    TYPECODE_MAP[vout32] = Type(VectorOut, 'f', n+1, True,  2, 1, True)
    TYPECODE_MAP[vout64] = Type(VectorOut, 'd', n+1, True,  6, 1, True)


def typecode_description(typecode):
    typeinfo = TYPECODE_MAP[typecode]
    name = typeinfo.devcls.__name__
    valuesize = calcsize(typeinfo.value_fmt)
    if typeinfo.num_params > 0 and not typeinfo.has_pctrl:
        name += f'/{typeinfo.num_params}'
    if typeinfo.num_values > 1:
        name += f'/{typeinfo.num_values}'
    if typecode in (0x1b03, 0x1f05):
        name += ' (32 bit, legacy)'
    else:
        name += f' ({valuesize*8} bit)'
    return name



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Support for an abstract communication protocol."""

from zapf import Error

ADDRESSES = '''Valid addresses:
ads://host[:port]/amsnetid:amsport
modbus://host[:port]/slaveno
tango://dbhost:dbport/tango/device/name'''


class Protocol:
    OFFSETS = []

    def __init__(self, url, log):
        self.url = url
        self.log = log
        self.offset = 0
        self.connected = False
        self._connect_callback = lambda _: None

    def set_callback(self, callback):
        # TODO for all protos: on hard comm error, set status to disconnected
        self._connect_callback = callback

    @staticmethod
    def create(url, log):
        if url.startswith('ads://'):
            return ADSProtocol(url, log)
        elif url.startswith('modbus://'):
            return ModbusProtocol(url, log)
        elif url.startswith('tango://'):
            return TangoProtocol(url, log)
        elif url.startswith('sim://'):
            return SimProtocol(url, log)
        raise Error(f'unsupported connection URL: {url}\n{ADDRESSES}')

    def connect(self):
        raise NotImplementedError

    def disconnect(self):
        raise NotImplementedError

    def reconnect(self):
        self.connect()

    def read(self, addr, length):
        raise NotImplementedError

    def write(self, addr, data):
        raise NotImplementedError


# pylint: disable=wrong-import-position
from zapf.proto.ads import ADSProtocol
from zapf.proto.modbus import ModbusProtocol
from zapf.proto.sim import SimProtocol
from zapf.proto.tango import TangoProtocol



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""A client library for the  PILS specification for SCADA-PLC interfaces."""


class Error(Exception):
    """Base for Zapf errors."""


class CommError(Error):
    """Raised for communication errors."""


class SpecError(Error):
    """Raised for invalid PLCs, or invalid data from the PLC."""


class PlcError(Error):
    """Raised for errors coming from the PLC."""


class ApiError(Error):
    """Raised for invalid API usage."""



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""IO abstraction."""

import threading
from struct import Struct, pack, unpack

from zapf import CommError
from zapf.indexer import Indexer
from zapf.proto import Protocol

# "Float order" converters.

def float_normal(x):
    return x


def float_swapped(x):
    return b''.join(x[i+2:i+4] + x[i:i+2] for i in range(0, len(x), 4))


class PlcIO:
    """IO caching object, one per PLC connection."""

    def __init__(self, proto, log):
        self.log = log
        if isinstance(proto, Protocol):
            self.proto = proto
        else:
            self.proto = Protocol.create(proto, log)
            self.proto.connect()
        self.cache = Cache(self.proto, log)
        self.magic = None
        self.indexer = Indexer(self, log)
        self.byteorder = '<'
        self.floatconv = float_normal

    def set_callback(self, callback):
        self.cache.callback = callback

    def set_update_params(self, cycle_secs, retries=-1):
        self.cache.cycle = cycle_secs
        self.cache.retries = retries

    def start_cache(self):
        self.cache.start()

    def stop_cache(self):
        self.cache.stop()

    # Cache related API.

    def register_cache_range(self, addr, length):
        """Register a cached area."""
        if length < 1:
            return
        self.cache.update_range(addr, length)
        self.sync_cache(wait=False)

    def sync_cache(self, wait=True):
        """To be called after a write to trigger a sync and wait for it"""
        if self.cache.running:
            self.cache.updated.clear()
            self.cache.trigger.set()
            if wait:
                self.cache.updated.wait(self.cache.cycle)

    # Higher level stuff.

    def detect_magic(self):
        if self.magic is not None:
            return self.magic

        def scan_offset(offset):
            for byteorder in '<>':
                self.byteorder = byteorder
                for floatconv in (float_normal, float_swapped):
                    self.floatconv = floatconv
                    try:
                        magic = self.read_f32(offset)
                    except CommError:
                        return False  # don't retry here
                    if 2014.0 <= magic < 2046.0:
                        byteorder = {'<': 'little', '>': 'big'}[self.byteorder]
                        floatconv = ', swapped floats' \
                            if self.floatconv is float_swapped else ''
                        self.log.info(f'found PLC according to {magic:.2f}: '
                                      f'{byteorder} endian{floatconv}')
                        self.magic = magic
                        self.proto.offset = offset
                        return True
            return False

        for offset in self.proto.OFFSETS:
            if scan_offset(offset):
                return self.magic
        raise CommError('did not find any MAGIC number')

    def float_from_dword(self, dw):
        return unpack('<f', self.floatconv(pack('<I', dw)))[0]

    # Read operations: cached accesses.

    def read_bytes(self, addr, num):
        return self.cache.get(addr, num)

    def read_fmt(self, addr, fmt):
        fmt = Struct(self.byteorder + fmt)
        return fmt.unpack(self.cache.get(addr, fmt.size))

    def read_u16(self, addr):
        return unpack(self.byteorder + 'H', self.cache.get(addr, 2))[0]

    def read_u16s(self, addr, num):
        return unpack(self.byteorder + str(num) + 'H',
                      self.cache.get(addr, 2*num))

    def read_i16(self, addr):
        return unpack(self.byteorder + 'h', self.cache.get(addr, 2))[0]

    def read_i16s(self, addr, num):
        return unpack(self.byteorder + str(num) + 'h',
                      self.cache.get(addr, 2*num))

    def read_u32(self, addr):
        return unpack(self.byteorder + 'I', self.cache.get(addr, 4))[0]

    def read_u32s(self, addr, num):
        return unpack(self.byteorder + str(num) + 'I',
                      self.cache.get(addr, 4*num))

    def read_i32(self, addr):
        return unpack(self.byteorder + 'i', self.cache.get(addr, 4))[0]

    def read_i32s(self, addr, num):
        return unpack(self.byteorder + str(num) + 'i',
                      self.cache.get(addr, 4*num))

    def read_u64(self, addr):
        return unpack(self.byteorder + 'Q', self.cache.get(addr, 8))[0]

    def read_u64s(self, addr, num):
        return unpack(self.byteorder + str(num) + 'Q',
                      self.cache.get(addr, 8*num))

    def read_i64(self, addr):
        return unpack(self.byteorder + 'q', self.cache.get(addr, 8))[0]

    def read_i64s(self, addr, num):
        return unpack(self.byteorder + str(num) + 'q',
                      self.cache.get(addr, 8*num))

    def read_f32(self, addr):
        return unpack(self.byteorder + 'f',
                      self.floatconv(self.cache.get(addr, 4)))[0]

    def read_f32s(self, addr, num):
        return unpack(self.byteorder + str(num) + 'f',
                      self.floatconv(self.cache.get(addr, 4*num)))

    # Note: floatconv is not used for 64-bit floats, since the PLCs that would
    # need floatconv don't support them.

    def read_f64(self, addr):
        return unpack(self.byteorder + 'd', self.cache.get(addr, 8))[0]

    def read_f64s(self, addr, num):
        return unpack(f'{self.byteorder}{num}d', self.cache.get(addr, 8*num))

    # Write operations: go direct, but trigger a cache sync afterwards.

    def write_i16(self, addr, data):
        self.write_bytes(addr, pack(self.byteorder + 'h', data))

    def write_u16(self, addr, data):
        self.write_bytes(addr, pack(self.byteorder + 'H', data))

    def write_i32(self, addr, data):
        self.write_bytes(addr, pack(self.byteorder + 'i', data))

    def write_u32(self, addr, data):
        self.write_bytes(addr, pack(self.byteorder + 'I', data))

    def write_i64(self, addr, data):
        self.write_bytes(addr, pack(self.byteorder + 'q', data))

    def write_u64(self, addr, data):
        self.write_bytes(addr, pack(self.byteorder + 'Q', data))

    def write_f32(self, addr, data):
        self.write_bytes(addr,
                         self.floatconv(pack(self.byteorder + 'f', data)))

    def write_f64(self, addr, data):
        self.write_bytes(addr, pack(self.byteorder + 'd', data))

    def write_u16_f32(self, addr, u16, f32):
        self.write_bytes(addr, pack(self.byteorder + 'H', u16) +
                         self.floatconv(pack(self.byteorder + 'f', f32)))

    def write_f32_u16(self, addr, f32, u16):
        self.write_bytes(addr, self.floatconv(pack(self.byteorder + 'f', f32)) +
                         pack(self.byteorder + 'H', u16))

    def write_f32_u32(self, addr, f32, u32):
        self.write_bytes(addr, self.floatconv(pack(self.byteorder + 'f', f32)) +
                         pack(self.byteorder + 'I', u32))

    def write_bytes(self, addr, data):
        with self.cache.lock:
            self.proto.write(addr, data)
        self.sync_cache()

    def write_fmt(self, addr, fmt, *data):
        with self.cache.lock:
            data = pack(self.byteorder + fmt, *data)
            self.proto.write(addr, data)
        self.sync_cache()

    def write_u16s(self, addr, data):
        packed = pack(f'{self.byteorder}{len(data)}H', *data)
        with self.cache.lock:
            self.proto.write(addr, packed)
        self.sync_cache()

    def write_f32s_u16(self, addr, data, u16):
        packed = b''
        for val in data:
            packed += self.floatconv(pack(f'{self.byteorder}f', val))
        packed += pack(f'{self.byteorder}H', u16)
        with self.cache.lock:
            self.proto.write(addr, packed)
        self.sync_cache()

    def write_f64s_u32(self, addr, data, u32):
        packed = pack(f'{self.byteorder}{len(data)}dI', *data, u32)
        with self.cache.lock:
            self.proto.write(addr, packed)
        self.sync_cache()


class Cache:
    """The actual cache."""

    def __init__(self, proto, log):
        self.log = log
        self.proto = proto
        self.cycle = 0.5
        self.retries = -1
        self.callback = lambda: None
        # data bytes
        self._data = None
        # range to cache
        self._range = (0, 0)   # empty
        # range that is *actually* cached
        self._cached_range = (0, 0)  # empty
        self.lock = threading.RLock()
        self.trigger = threading.Event()  # trigger
        self.updated = threading.Event()  # ack
        self._retried = 0
        self.running = False
        self._thread = None

    def start(self):
        if self._thread and self._thread.is_alive():
            return
        self.running = True
        self._thread = threading.Thread(target=self.run, daemon=True,
                                        name=f'Caching {self.proto.url}')
        self._thread.start()

    def stop(self):
        if self._thread:
            self.running = False
            self._thread.join()

    def update_range(self, addr, length):
        with self.lock:
            # first area to cache?
            if self._range[0] == self._range[1]:
                first = addr
                above = addr + length
            else:
                first = min(self._range[0], addr)
                above = max(self._range[1], addr + length)
            # cache range stores (start, end+1)
            self._range = first, above

    def get(self, addr, length):
        with self.lock:
            if addr >= self._cached_range[0] and \
               addr + length <= self._cached_range[1]:
                addr -= self._cached_range[0]
                return self._data[addr:addr + length]
            return self.proto.read(addr, length)

    def run(self):
        """Loop which updates the cache whenever cycle seconds have passed
        after the last update, or trigger was set.
        """
        while self.running:
            self.trigger.wait(self.cycle)
            self.trigger.clear()
            self.sync()

    def sync(self):
        """Update our cache and set the cache.updated event."""
        with self.lock:
            if self._range != (0, 0):
                try:
                    cached_data = self.proto.read(
                        self._range[0], self._range[1] - self._range[0])
                    if self._data is None:
                        self.log.info('initial cache update ok')
                    self._retried = 0
                except Exception:
                    if self._data:
                        self.log.warning('problem while trying to sync cache')
                    self._data = None
                    self._retried += 1
                    if self._retried == self.retries:
                        self.log.info('cache sync aborted')
                        self.running = False
                        self.updated.set()
                        return
                else:
                    self._cached_range = self._range
                    self._data = cached_data
                    self.callback()
        self.updated.set()



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Implements communication via Modbus/TCP."""

import re
import socket
from struct import Struct, pack, unpack, unpack_from

from zapf import ApiError, CommError
from zapf.proto import Protocol

MB_ADDR_RE = re.compile(r'modbus://(.+?)(:(\d+))?(?:/(\d+)?)?$')

DEFAULT_PORT = 502

FUNC_READ_HOLDING = 3
FUNC_WRITE_MULTI = 16

HEADER = Struct('>H2xHBB')
HEADER_SIZE = HEADER.size
TWO_WORDS = Struct('>HH')


class ModbusProtocol(Protocol):
    OFFSETS = [0, 0x6000, 0x8000]

    def __init__(self, url, log):
        adr = MB_ADDR_RE.match(url)
        if not adr:
            raise ApiError('invalid Modbus address, must be '
                           'modbus://host[:port]/slave')
        host = adr.group(1)
        if adr.group(2):
            port = int(adr.group(3))
        else:
            port = DEFAULT_PORT
        self._slave = int(adr.group(4) or 0)
        self._iphostport = (host, port)
        self._socket = None
        self._trans_id = 1

        Protocol.__init__(self, url, log)

    def connect(self):
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self._socket.connect(self._iphostport)
            self.log.info('connected to %s', self.url)
        except (OSError, ValueError) as err:
            raise CommError(f'could not connect to {self.url}: {err}') from err
        self._connect_callback(True)
        self.connected = True

    def disconnect(self):
        try:
            self._socket.shutdown(socket.SHUT_RDWR)
            self._socket.close()
        except OSError:
            pass
        self._socket = None
        self._connect_callback(False)
        self.connected = False

    def read(self, addr, length):
        if not self.connected:
            self.reconnect()
        assert addr % 2 == 0
        nregs = (length + 1) // 2
        payload = TWO_WORDS.pack((self.offset + addr) // 2, nregs)
        try:
            reply = self._comm(FUNC_READ_HOLDING, payload, 2*nregs + 1)
        except OSError as err:
            self.log.exception('during read')
            self.disconnect()
            raise CommError('IO error during read: %s' % err) from err
        reply_regs = unpack_from('>%dH' % nregs, reply, 1)
        return pack('<%dH' % nregs, *reply_regs)[:length]

    def write(self, addr, data):
        if not self.connected:
            self.reconnect()
        # odd addrs and data lengths should never be needed since the
        # protocol is defined with 16-bit units in mind
        assert addr % 2 == 0 and len(data) % 2 == 0
        nregs = len(data) // 2
        payload = pack('>HHB%dH' % nregs, (self.offset + addr) // 2,
                       nregs, len(data), *unpack('<%dH' % nregs, data))
        try:
            reply = self._comm(FUNC_WRITE_MULTI, payload, 4)
        except OSError as err:
            self.log.exception('during write')
            self.disconnect()
            raise CommError('IO error during write: %s' % err) from err
        if reply != payload[:4]:
            raise CommError('invalid data in write reply')

    def _comm(self, function, payload, exp_len):
        self._trans_id = (self._trans_id + 1) & 0xFFFF
        msg = HEADER.pack(self._trans_id, len(payload) + 2,
                          self._slave, function) + payload
        expected = HEADER_SIZE + exp_len
        self._socket.sendall(msg)
        reply = b''
        rephdr = None
        while len(reply) < expected:
            data = self._socket.recv(expected - len(reply))
            if not data:
                raise CommError('no data in read')
            reply += data
            if len(reply) >= HEADER_SIZE and not rephdr:
                rephdr = HEADER.unpack_from(reply)
                if rephdr[0] != self._trans_id:
                    raise CommError('invalid transaction ID in reply')
                if rephdr[2] != self._slave:
                    raise CommError('invalid slave ID in reply')
                if rephdr[3] != function:
                    self._translate_error(reply[HEADER_SIZE:])
                if rephdr[1] != exp_len + 2:
                    raise CommError('unexpected length in reply')
        if rephdr[1] != len(reply) - 6:
            raise CommError('invalid length in reply or packet truncated')
        return reply[HEADER_SIZE:]

    def _translate_error(self, data):
        raise CommError('modbus: ' +
                        EXC_STRINGS.get(data[0], 'exception %d' % data[0]))


EXC_STRINGS = {
    1: 'illegal function',
    2: 'illegal data address',
    3: 'illegal data value',
    4: 'slave device failure',
    5: 'acknowledge',
    6: 'slave device busy',
    7: 'negative acknowledge',
    8: 'memory parity error',
    10: 'gateway path unavailable',
    11: 'gateway target device failed to respond',
}



#  -*- coding: utf-8 -*-
# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

from zapf.simulator.runtime import mem


def adr(val):
    return val.addr


def sizeof(val):
    return val.sizeof()


def memset(addr, byte, count):
    mem.write(addr, chr(byte).encode() * count)


def memcpy(toaddr, fromaddr, count):
    data = mem.read(fromaddr, count)
    mem.write(toaddr, data)



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Contains all necessary constants and mappings from the spec."""

from zapf.util import BitFields, ParamMap, UncasedMap

SUPPORTED_MAGICS = ['2015_02']

OFFSET_ADDR = 4

INDEXER_DEV = 0

INFO_STRUCT  = 0
INFO_SIZE    = 1
INFO_ADDR    = 2
INFO_UNIT    = 3
INFO_NAME    = 4
INFO_VERSION = 5
INFO_AUTHOR1 = 6
INFO_AUTHOR2 = 7
INFO_PARAMS  = 15
INFO_AUX1    = 16
INFO_CYCLE   = 127


UNIT_CODES    = ('', 'V', 'A', 'W', 'm', 'g', 'Hz', 'T', 'K', 'degC', 'degF',
                 'bar', 'deg', 'Ohm', 'm/s', 'm^2/s', 'm^3/s', 's', 'cts',
                 'bar/s', 'bar/s^2', 'F', 'H', 'l/min')
UNIT_EXPONENT = {0: '', 2: 'h', 3: 'k', 6: 'M', 9: 'G', 12: 'T', 15: 'P',
                 18: 'E', -1: 'd', -2: 'c', -3: 'm', -6: 'u', -9: 'n',
                 -12: 'f', -15: 'a'}
UNIT_SPECIAL  = {(-2, 0): '%', (-3, 16): 'l/s'}

FLOAT32_MAX = 3.402823e+38
FLOAT64_MAX = 1.7976931348623158e+308

StatusStruct = BitFields(STATE=(15, 12), REASON=(11, 8), AUX=(7, 0))
ParamControl = BitFields(available=(15,15), CMD=(15, 13), SUBINDEX=(12, 8), IDX=(7, 0))

ReasonMap = [
    '',
    'inhibit',
    'timeout',
    '(inhibit, timeout)',
    'lower limit reached',
    '(inhibit, lower limit)',
    '(timeout, lower limit)',
    '(inhibit, timeout, lower limit)',
    'upper limit reached',
    '(inhibit, upper limit)',
    '(timeout, upper limit)',
    '(inhibit, timeout, upper limit)',
    'both limits reached',
    '(inhibit, both limits)',
    '(timeout, both limits)',
    '(inhibit, timeout, both limits)',
]

ParamCMDs = UncasedMap(
    INIT=0,
    DO_READ=1,
    DO_WRITE=2,
    BUSY=3,
    DONE=4,
    ERR_NO_IDX=5,
    ERR_RO=6,
    ERR_RETRY=7,
)

PLCStatus = UncasedMap(
    RESET=0,
    IDLE=1,
    DISABLED=2,
    WARN=3,
    START=5,
    BUSY=6,
    STOP=7,
    ERROR=8,
    DIAGNOSTIC_ERROR=13,
)

# The first (numbered) parameter whose value is float instead of integer.
FIRST_FLOAT_PARAM = 30

# Parameter 0 should not be used.
Parameters = ParamMap(
    UNUSED=0,
    Mode=1,
    Microsteps=2,
    CoderBits=3,
    ExtStatus=4,
    # HwOffset=5,
    _AbsMin=30,  # deprecated
    _AbsMax=31,  # deprecated
    UserMin=32,
    UserMax=33,
    WarnMin=34,
    WarnMax=35,
    TimeoutTime=36,
    MaxTravelDist=37,
    AccelTime=38,
    Offset=40,
    BlockSize=43,
    Opening=44,
    PidP=51,
    PidI=52,
    PidD=53,
    DragError=55,
    Hysteresis=56,
    Holdback=57,
    HomingSpeed=58,
    Jerk=59,
    Speed=60,
    Accel=61,
    IdleCurrent=62,
    RampCurrent=63,
    MoveCurrent=64,
    StopCurrent=65,
    # _AbortDecel=66,  # deprecated
    _Microsteps=67,  # deprecated
    Slope=68,
    HomePosition=69,
    Setpoint=70,

    # special functions

    FactoryReset=128,
    _RefMinus=131,  # deprecated
    _RefPlus=132,  # deprecated
    Home=133,
    SetPosition=137,
    ContMove=142,
)


def is_function(idx):
    return 128 <= idx < 192 or 224 <= idx < 240



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Support for scanning the indexer and creating devices."""

from collections import namedtuple

from zapf import spec
from zapf.device import Device
from zapf.io import PlcIO

DeviceInfo = namedtuple('DeviceInfo', 'number name addr typecode info')


class Scanner:
    def __init__(self, io_or_proto, log):
        self.log = log
        if isinstance(io_or_proto, PlcIO):
            self.io = io_or_proto
        else:
            self.io = PlcIO(io_or_proto, log)

    def get_plc_data(self):
        """Establish communication with the indexer and return it."""
        self.io.indexer.detect_plc()
        return self.io.indexer

    def get_devices(self, _device_filter=None):
        """Scan the PLC and yield device objects."""
        for data in self.scan_devices():
            devcls = Device.class_for(data.typecode)
            if devcls is None:
                self.log.warning('type code %#x is not supported, skipping',
                                 data.typecode)
            else:
                yield devcls(data.number, data.name, data.addr, data.typecode,
                             data.info, self.io, self.log)

    def scan_devices(self):
        """Scan the PLC and yield information about devices."""

        # ensure we can talk to the indexer
        self.io.indexer.detect_plc()

        # check which method we need to scan
        method = getattr(self, 'scan_' + self.io.indexer.magicstr, None)
        if not method:
            raise RuntimeError('Magic %s is supported, but no scanner method '
                               'available' % self.io.indexer.magicstr)

        yield from method()

    def scan_2015_02(self):
        indexer = self.io.indexer
        next_addr = indexer.indexer_addr + indexer.indexer_size

        for devnum in range(1, 256):
            info = indexer.query_infostruct(devnum)
            typecode, size, addr, unit, flags, absmin, absmax, name = info

            # gone past last device?
            if typecode == 0:
                break

            # if there is not valid data in the infostruct, query individually
            if size + addr + flags == 0 and not name:
                size = indexer.query_word(devnum, spec.INFO_SIZE)
                if not size:
                    size = 2 * (typecode & 0xff)
                addr = indexer.query_word(devnum, spec.INFO_ADDR)
                if not addr:
                    addr = next_addr
                unit = indexer.query_unit(devnum, spec.INFO_UNIT)
                flags = None
                lowlevel = False
                absmin = -spec.FLOAT32_MAX
                absmax = spec.FLOAT32_MAX
            else:
                lowlevel = (flags & 0x80000000) != 0

            # this might be empty even if a valid infostruct was present,
            # e.g. if the full name didn't fit behind the 20 previous bytes
            if not name:
                name = indexer.query_string(devnum, spec.INFO_NAME)

            # extract info about parameters and special functions
            devclass = typecode >> 13
            parameters = []
            functions = []
            if devclass in (1, 2):
                if devclass == 1:  # FlatDevices: a list of indices
                    plist = indexer.query_bytes(devnum, spec.INFO_PARAMS)
                    param_ids = [p for p in plist if p]
                else:  # ParamDevices: a bitmap of indices
                    param_ids = indexer.query_bitmap(devnum, spec.INFO_PARAMS)
                param_ids.sort()  # ... which *should* be unnecessary
                for p in param_ids:
                    if spec.is_function(p):
                        functions.append(spec.Parameters[p])
                    else:
                        parameters.append(spec.Parameters[p])

            # extract AUX string labels
            aux_strings = [''] * 24
            if flags is None:  # old way, stop reading the first empty one
                for idx in range(24):
                    aux_string = indexer.query_string(devnum,
                                                      spec.INFO_AUX1 + idx)
                    aux_strings[idx] = aux_string.strip()
                    if not aux_string:
                        break
            else:  # new way, only read the relevant (flagged ones)
                for idx in range(24):
                    if flags & (1 << idx):
                        aux_string = indexer.query_string(devnum,
                                                          spec.INFO_AUX1 + idx)
                        aux_strings[idx] = aux_string.strip()

            info = dict(lowlevel=lowlevel, unit=unit,
                        absmin=absmin, absmax=absmax,
                        params=parameters, funcs=functions,
                        aux_strings=aux_strings)
            self.log.info('found device %s at addr %s with type %#x',
                          name, addr, typecode)
            yield DeviceInfo(devnum, name, addr, typecode, info)

            next_addr = addr + size



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Implements communication via a Tango device."""

import re
from struct import pack, unpack

from zapf import ApiError, CommError, Error
from zapf.proto import Protocol

try:
    import PyTango
except ImportError:
    PyTango = None

TANGO_ADDR_RE = tangodev_re = re.compile(
    r'tango://([\w.-]+:[\d]+/)?([\w-]+/){2}[\w-]+(#dbase=(no|yes))?$')


class TangoProtocol(Protocol):
    # Since a Tango device can represent both Modbus and ADS, we have to try
    # all offsets.
    OFFSETS = [0, 0x6000, 0x8000]

    def __init__(self, url, log):
        if not PyTango:
            raise Error('install PyTango to use the tango:// protcol')

        if not TANGO_ADDR_RE.match(url):
            raise ApiError('invalid Tango address, must be '
                           'ads://host/amsnetid:amsport')
        self._proxy = None

        Protocol.__init__(self, url, log)

    def connect(self):
        try:
            self._proxy = PyTango.DeviceProxy(self.url)
            self._proxy.state
        except PyTango.DevFailed as err:
            raise CommError('cannot connect to Tango device: %s' % err) from err
        except AttributeError:
            raise CommError('Tango device exists, but seems to be not '
                            'running') from None
        try:
            self._proxy.ReadInputBytes
            self._proxy.WriteOutputBytes
            self.read = self._read_ads
            self.write = self._write_ads
        except AttributeError:
            try:
                self._proxy.ReadOutputWords
                self._proxy.WriteOutputWords
                self.read = self._read_modbus
                self.write = self._write_modbus
            except AttributeError:
                raise CommError('Tango device seems to have the '
                                'wrong interface (needs to be Profibus '
                                'or Modbus)') from None
        self._connect_callback(True)
        self.connected = True

    def disconnect(self):
        self._proxy = None
        self._connect_callback(False)
        self.connected = False

    # These methods are overwritten by the protocol-specific ones below.

    def read(self, addr, length):
        raise ApiError('not connected')

    def write(self, addr, data):
        raise ApiError('not connected')

    # I/O using an ADS Tango device ("Profibus" Entangle interface)

    def _read_ads(self, addr, length):
        if not self.connected:
            self.reconnect()
        try:
            result = self._proxy.ReadInputBytes((addr, length))  # no offset
        except PyTango.DevFailed as err:
            self.log.exception('during read')
            raise CommError('Tango error during read: %s' % err) from err
        return bytes(result)

    def _write_ads(self, addr, data):
        if not self.connected:
            self.reconnect()
        try:
            self._proxy.WriteOutputBytes(((addr,) + tuple(data)))
        except PyTango.DevFailed as err:
            self.log.exception('during read')
            raise CommError('Tango error during read: %s' % err) from err

    # I/O using a Modbus Tango device

    def _read_modbus(self, addr, length):
        if not self.connected:
            self.reconnect()
        assert addr % 2 == 0
        nregs = (length + 1) // 2
        try:
            result = self._proxy.ReadOutputWords(((self.offset + addr) // 2,
                                                  nregs))
        except PyTango.DevFailed as err:
            self.log.exception('during read')
            raise CommError('Tango error during read: %s' % err) from err
        return pack('<%dH' % nregs, *result)[:length]

    def _write_modbus(self, addr, data):
        if not self.connected:
            self.reconnect()
        assert addr % 2 == 0 and len(data) % 2 == 0
        nregs = len(data) // 2
        payload = ((self.offset + addr) // 2,) + unpack('<%dH' % nregs, data)
        try:
            self._proxy.WriteOutputWords(payload)
        except PyTango.DevFailed as err:
            self.log.exception('during read')
            raise CommError('Tango error during read: %s' % err) from err



#  -*- coding: utf-8 -*-
# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

import logging
import struct

import pytest

from zapf import CommError
from zapf.device import TYPECODE_MAP, typecode_description
from zapf.io import PlcIO
from zapf.proto.sim import SimProtocol
from zapf.scan import Scanner

from test import testplc


def prettify(data):
    if len(data) == 2:
        val = struct.unpack('H', data)[0]
        return f'{val} {val:#x}'
    elif len(data) == 4:
        val_i = struct.unpack('I', data)[0]
        val_f = struct.unpack('f', data)[0]
        return f'{val_i} {val_i:#x} {val_f}'
    elif len(data) == 8:
        val_i = struct.unpack('Q', data)[0]
        val_f = struct.unpack('d', data)[0]
        return f'{val_i} {val_i:#x} {val_f}'
    return ''


class TestProtocol(SimProtocol):
    OFFSETS = [0, 0x10000]  # allow trying a failing offset

    def read(self, addr, length):
        try:
            data = super().read(addr, length)
        except Exception as err:
            self.log.warning(f'R {addr:#x} {length} !! {err}')
            raise CommError(f'read failed: {err}') from err
        else:
            self.log.debug(f'R {addr:#x} {length} -> {data} {prettify(data)}')
            return data

    def write(self, addr, data):
        try:
            super().write(addr, data)
        except Exception as err:
            self.log.warning(f'W {addr:#x} {data} {prettify(data)} !! {err}')
            raise CommError(f'write failed: {err}') from err
        else:
            self.log.debug(f'W {addr:#x} {data} {prettify(data)} ok')


@pytest.fixture(scope='module')
def plc_io():
    proto = TestProtocol(testplc.Main, logging.getLogger('simplc'))
    proto.connect()

    yield PlcIO(proto, logging.root)

    proto.disconnect()


# preselect implemented typecodes from BIG TABLE
tcm = [(t, dinfo) for (t, dinfo) in TYPECODE_MAP.items()
       if dinfo.num_values in [0, 1, 2, 8, 16]
       if dinfo.num_params in [0, 1, 8, 16]]

# slice defined typcodes into smaller sets for the specific tests
# These slices are tied to the corresponding tests in test_device
# we split into:
# * readable devices (single value) (contains devices with target too),
# * devices with a target (single value)
# * devices with flat parameters
# * devices with param interface
# * vector devices
TC_READABLE_DEVICES = [t for (t, dinfo) in tcm if dinfo.num_values == 1]
TC_TARGET_DEVICES = [t for (t, dinfo) in tcm if dinfo.num_values == 1
                     if (dinfo.has_target or (t >> 8) in [0x14, 0x15])]
TC_FLAT_DEVICES = [t for (t, dinfo) in tcm if dinfo.num_values == 1
                   if dinfo.num_params >= 1 if not dinfo.has_pctrl]
TC_PARAM_DEVICES = [t for (t, dinfo) in tcm if dinfo.num_params >= 1
                    if dinfo.has_pctrl]
TC_VECTOR_DEVICES = [t for (t, dinfo) in tcm if dinfo.num_values >= 2]


# Scan testplc only once for implemented devices
CACHED_DEVLIST = []


# pylint: disable=redefined-outer-name, inconsistent-return-statements
def filter_tc(plc_io, tc):
    """find a zapf-device with the requested typecode"""
    if not CACHED_DEVLIST:
        CACHED_DEVLIST[:] = Scanner(plc_io, plc_io.log).get_devices()
    # find the dev with typecode
    for d in CACHED_DEVLIST:
        if d.typecode == tc:
            return d, TYPECODE_MAP[tc]
    pytest.fail(f'typecode 0x{tc:04x} not implemented in testplc')


@pytest.fixture(scope='module', params=TC_READABLE_DEVICES,
                ids=typecode_description)
# pylint: disable=redefined-outer-name
def plc_readable_device(plc_io, request):
    return filter_tc(plc_io, request.param)


@pytest.fixture(scope='module', params=TC_TARGET_DEVICES,
                ids=typecode_description)
# pylint: disable=redefined-outer-name
def plc_target_device(plc_io, request):
    return filter_tc(plc_io, request.param)


@pytest.fixture(scope='module', params=TC_FLAT_DEVICES,
                ids=typecode_description)
# pylint: disable=redefined-outer-name
def plc_flat_device(plc_io, request):
    return filter_tc(plc_io, request.param)


@pytest.fixture(scope='module', params=TC_PARAM_DEVICES,
                ids=typecode_description)
# pylint: disable=redefined-outer-name
def plc_param_device(plc_io, request):
    return filter_tc(plc_io, request.param)


@pytest.fixture(scope='module', params=TC_VECTOR_DEVICES,
                ids=typecode_description)
# pylint: disable=redefined-outer-name
def plc_vector_device(plc_io, request):
    return filter_tc(plc_io, request.param)


@pytest.fixture(scope='function', autouse=True)
def log_fixture(caplog):
    caplog.set_level(logging.INFO)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

from setuptools import find_packages, setup

import zapf.version

pkgdata = ['RELEASE-VERSION']

setup(
    name = 'zapf',
    version = zapf.version.get_version(),
    license = 'GPL-2.0+',
    author = 'Georg Brandl',
    author_email = 'g.brandl@fz-juelich.de',
    description = 'Client library for the PILS specification',
    packages = find_packages(),
    package_data = {'zapf': pkgdata},
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Human Machine Interfaces',
    ],
)



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Basic test suite for the scanner."""

from zapf.device import typecode_description

from zapf.scan import Scanner


def test_scan(plc_io):
    scanner = Scanner(plc_io, plc_io.log)

    plc_data = scanner.get_plc_data()
    assert plc_data.indexer_addr == 64
    assert plc_data.indexer_size == 36
    assert plc_data.plc_name == 'lazy test plc'
    assert plc_data.plc_author == 'anonymous\ncoward'
    assert plc_data.plc_version == '0.0.1alpha'

    for devinfo in scanner.scan_devices():
        # the test devices have the same name as the typecode description
        assert typecode_description(devinfo.typecode) == devinfo.name
        # make sure we get all required keys in the info dict
        for key in ('lowlevel', 'unit', 'absmin', 'absmax',
                    'params', 'funcs', 'aux_strings'):
            assert key in devinfo.info
